@@ -122,8 +122,8 @@ int
 acctswon(void)
 {
 	int			i, j, sematopid;
-	static ushort 		vals[] = {1, ATOPACCTTOT};
-	union {ushort *array;}	arg = {vals};
+	static unsigned short 		vals[] = {1, ATOPACCTTOT};
+	union {unsigned short *array;}	arg = {vals};
 	struct stat		statbuf;
 	char			*ep;
 	int			ret;
@@ -430,7 +430,6 @@ atopacctd(int swon)
 				{
 					if ( swon && !acctvers(acctfd) )
 					{
-
 						int maxcnt = 40;
 
 						if ( fork() == 0 )
@@ -453,7 +452,8 @@ atopacctd(int swon)
 								&semunlock, 1);
 
 							regainrootprivs();
-							return 1;
+							maxshadowrec = 0;
+							return -1;  // try other
 						}
 					}
 
