@@ -109,8 +109,11 @@
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <signal.h>
 #include <regex.h>
 #include <sys/utsname.h>
+#include <sys/select.h>
+#include <sys/ioctl.h>
 
 #include "atop.h"
 #include "photoproc.h"
@@ -118,7 +121,8 @@
 #include "photosyst.h"
 
 extern char usecolors;
-
+static char winchange;
+static char keywaiting;	// key pushed back in stream?
 
 // maximum X axis label length
 //
@@ -229,6 +233,7 @@ static int	drawmemlines(struct perwindow *, int, int, int, int,
 static int	drawevent(struct perwindow *, int, int, int,
 			char *, char *, long);
 
+static void	getsigwinch(int);
 
 /////////////////////////////////////////////////////
 // entry point to display the deviation counters
@@ -344,7 +349,7 @@ draw_samp(time_t curtime, int nsecs, struct sstat *sstat,
 				statuscol = FGCOLORCRIT;
 		}
 
-		// wait for keystroke or alarm expiration
+		// wait for keystroke, alarm expiration or inotify watcher
 		//
 		switch (lastchar = headergetch(curtime, nsecs, statusmsg, statuscol))
 		{
@@ -352,6 +357,7 @@ draw_samp(time_t curtime, int nsecs, struct sstat *sstat,
 		   case 0:
 			return lastchar;
 
+		   case MCGROUPS:
 		   case MPROCGEN:
 		   case MPROCMEM:
 		   case MPROCDSK:
@@ -387,7 +393,7 @@ draw_samp(time_t curtime, int nsecs, struct sstat *sstat,
 			break;		// redraw with current values
 
 		   case MSAMPNEXT:	// manual trigger for next sample?
-			if (paused)
+			if (paused && !twinpid)
 				break;
 
 			getalarm(0);
@@ -400,8 +406,12 @@ draw_samp(time_t curtime, int nsecs, struct sstat *sstat,
 				break;
 			}
 
-			if (paused)
-				break;
+			if (!paused && twinpid)
+			{
+				paused=1;       // implicit pause in twin mode
+				clrtoeol();
+				refresh();
+			}
 
 			getalarm(0);
 			return lastchar;
@@ -409,10 +419,18 @@ draw_samp(time_t curtime, int nsecs, struct sstat *sstat,
 		   case MRESET:		// reset to begin?
 			getalarm(0);
 			paused = 0;
+
+			if (twinpid)
+			{
+				paused=1;       // implicit pause in twin mode
+				clrtoeol();
+				refresh();
+			}
+
 			return lastchar;
 
 		   case MSAMPBRANCH:	// branch to other time?
-			// only possible when viewing raw file
+			// only possible in twin mode or when viewing raw file
 			//
 			if (!rawreadflag)
 			{
@@ -420,6 +438,13 @@ draw_samp(time_t curtime, int nsecs, struct sstat *sstat,
 				break;
 			}
 
+			if (!paused && twinpid)
+			{
+				paused=1;       // implicit pause in twin mode
+				clrtoeol();
+				refresh();
+			}
+
 			if (getwininput("Enter new time (format [YYYYMMDD]hhmm): ",
 					answer, sizeof answer, 1) >= 0)
 			{
@@ -449,12 +474,25 @@ draw_samp(time_t curtime, int nsecs, struct sstat *sstat,
 			break;
 
 		   case MPAUSE:		// pause key (toggle)?
+			if (rawreadflag && !twinpid)
+			{
+				beep();
+				break;
+			}
+
 			if (paused)
 			{
 				paused=0;
 
 				if (!rawreadflag)
+				{
 					alarm(1);
+				}
+				else
+				{
+					begintime = 0x7fffffff;
+					return MSAMPBRANCH;
+				}
 			}
 			else
 			{
@@ -464,7 +502,7 @@ draw_samp(time_t curtime, int nsecs, struct sstat *sstat,
 			break;
 
 		   case MINTERVAL:	// modify interval?
-			// not possible when viewing raw file
+			// not possible in twin mode or when viewing raw file
 			//
 			if (rawreadflag)
 			{
@@ -506,6 +544,13 @@ draw_samp(time_t curtime, int nsecs, struct sstat *sstat,
 			if (interval && !paused && !rawreadflag)
 				alarm(1); // force new sample
 
+			if (twinpid)    // twin mode?
+			{
+				// jump to last sample after awaiting input
+				begintime = 0x7fffffff;
+				return MSAMPBRANCH;
+			}
+
 			break;
 
 		   default:		// any other key pressed?
@@ -594,24 +639,24 @@ do_cpubars(struct sstat *sstat, int nsecs, char initlabels, char mono)
 
 	if (!mono)
 	{
-		vertvals->category[0].ccol	= COLORCPUSYS;
-		vertvals->category[0].clab	= 'S';
-		vertvals->category[0].cval	= sstat->cpu.all.stime * 100 /
+		vertvals->category[0].ccol	= COLORCPUINTR;
+		vertvals->category[0].clab	= 'I';
+		vertvals->category[0].cval	=
+			(sstat->cpu.all.Stime + sstat->cpu.all.Itime) * 100 /
+						  alltics;
+
+		vertvals->category[1].ccol	= COLORCPUSYS;
+		vertvals->category[1].clab	= 'S';
+		vertvals->category[1].cval	= sstat->cpu.all.stime * 100 /
 		 				  alltics;
 	
-		vertvals->category[1].ccol	= COLORCPUUSR;
-		vertvals->category[1].clab	= 'U';
-		vertvals->category[1].cval	=
+		vertvals->category[2].ccol	= COLORCPUUSR;
+		vertvals->category[2].clab	= 'U';
+		vertvals->category[2].cval	=
 			(sstat->cpu.all.utime +
 			 sstat->cpu.all.ntime -
 			 sstat->cpu.all.guest) * 100 / alltics;
 
-		vertvals->category[2].ccol	= COLORCPUIDLE;
-		vertvals->category[2].clab	= 'I';
-		vertvals->category[2].cval	=
-			(sstat->cpu.all.Stime + sstat->cpu.all.Itime) * 100 /
-						  alltics;
-
 		vertvals->category[3].ccol	= COLORCPUSTEAL;
 		vertvals->category[3].clab	= 's';
 		vertvals->category[3].cval	=
@@ -659,25 +704,25 @@ do_cpubars(struct sstat *sstat, int nsecs, char initlabels, char mono)
 
 			if (!mono)
 			{
-				(vertvals+i+1)->category[0].ccol = COLORCPUSYS;
-				(vertvals+i+1)->category[0].clab = 'S';
+				(vertvals+i+1)->category[0].ccol = COLORCPUINTR;
+				(vertvals+i+1)->category[0].clab = 'I';
 				(vertvals+i+1)->category[0].cval =
+					(sstat->cpu.cpu[i].Stime +
+					 sstat->cpu.cpu[i].Itime) *100/alltics;
+
+				(vertvals+i+1)->category[1].ccol = COLORCPUSYS;
+				(vertvals+i+1)->category[1].clab = 'S';
+				(vertvals+i+1)->category[1].cval =
 					sstat->cpu.cpu[i].stime * 100 /
 								alltics;
 
-				(vertvals+i+1)->category[1].ccol = COLORCPUUSR;
-				(vertvals+i+1)->category[1].clab = 'U';
-				(vertvals+i+1)->category[1].cval =
+				(vertvals+i+1)->category[2].ccol = COLORCPUUSR;
+				(vertvals+i+1)->category[2].clab = 'U';
+				(vertvals+i+1)->category[2].cval =
 					(sstat->cpu.cpu[i].utime +
 				 	 sstat->cpu.cpu[i].ntime -
 				 	 sstat->cpu.cpu[i].guest) *100/alltics;
 
-				(vertvals+i+1)->category[2].ccol = COLORCPUIDLE;
-				(vertvals+i+1)->category[2].clab = 'I';
-				(vertvals+i+1)->category[2].cval =
-					(sstat->cpu.cpu[i].Stime +
-					 sstat->cpu.cpu[i].Itime) *100/alltics;
-
 				(vertvals+i+1)->category[3].ccol = COLORCPUSTEAL;
 				(vertvals+i+1)->category[3].clab = 's';
 				(vertvals+i+1)->category[3].cval =
@@ -1466,7 +1511,7 @@ fillbarmaps(int numbars, struct vertval *vvp, float valperunit, int barlines)
 			//
 			n = (vp->category[c].cval + valperunit/2) / valperunit;
 
-			if (n > 0 && mp - vp->barmap + n < MAXHEIGHT)
+			if (n > 0 && mp - vp->barmap + n <= MAXHEIGHT)
 			{
 				// fill all color positions (lines)
 				//
@@ -2113,6 +2158,11 @@ headergetch(time_t curtime, int nsecs, char *statusmsg, int statuscol)
 	int	headlen = strlen(headmsg);
 	char	buf[64], timestr[16], datestr[16];
 	int	seclen  = val2elapstr(nsecs, buf);
+	int	lastchar;
+
+	fd_set	readfds;
+	char	eventbuf[1024];
+	int	nrfds;
 
 	convdate(curtime, datestr);       /* date to ascii string   */
 	convtime(curtime, timestr);       /* time to ascii string   */
@@ -2149,11 +2199,95 @@ headergetch(time_t curtime, int nsecs, char *statusmsg, int statuscol)
         	colorswoff(headwin, statuscol);
 	}
 
-	wrefresh(headwin);
-
 	// wait for keystroke
 	//
-	return mvwgetch(headwin, 1, 0);
+	wmove(headwin, 1, 0);
+
+	wrefresh(headwin);
+
+	if (twinpid && !keywaiting)	// twin mode?
+	{
+		struct sigaction sigact, sigold;
+
+		/*
+		** catch window size changes while in select
+		*/
+	        memset(&sigact, 0, sizeof sigact);
+       		sigact.sa_handler = getsigwinch;
+       		sigaction(SIGWINCH, &sigact, &sigold);
+
+		winchange = 0;
+
+		/*
+		** await input character from keyboard, or
+		**       inotify trigger in case of twin mode, or
+		**       interval timer expiration
+		*/
+		FD_ZERO(&readfds);
+		FD_SET(0, &readfds);
+
+		if (!paused && fdinotify != -1)	// twin mode?
+		{
+			FD_SET(fdinotify, &readfds);
+			nrfds = fdinotify + 1;
+		}
+		else
+		{
+			nrfds = 1;
+		}
+
+		switch (select(nrfds, &readfds, (fd_set *)0,
+	                      (fd_set *)0, (struct timeval *)0))
+		{
+		   case -1:
+			/*
+			** window change or timer expiration?
+			*/
+			if (winchange)
+			{
+				// window change: set new dimensions
+				struct winsize w;
+
+				ioctl(0, TIOCGWINSZ, &w);
+				resizeterm(w.ws_row, w.ws_col);
+				lastchar = KEY_RESIZE;
+			}
+			else
+			{
+				// time interrupt
+				lastchar = 0;
+			}
+			break;
+
+		   default:
+			/*
+			** inotify trigger that new sample has been written?
+			** pretend as if the 't' key has been pressed
+			** to read that sample
+			**
+			** otherwise: read keystroke from keyboard
+			*/
+			if (FD_ISSET(fdinotify, &readfds))
+			{
+				read(fdinotify, eventbuf, sizeof eventbuf);
+
+				lastchar = MSAMPNEXT;
+			}
+			else
+			{
+				lastchar = wgetch(headwin);
+				keywaiting = 0;
+			}
+		}
+
+      			sigaction(SIGWINCH, &sigold, (struct sigaction *)0);
+	}
+	else	// no twin mode: neutral state is getch()
+	{
+		lastchar = wgetch(headwin);
+	}
+
+	return lastchar;
 }
 
 
@@ -2446,6 +2580,8 @@ showhelp(void)
 	mvwprintw(helpwin, line++, 2, "Display mode:");
         mvwprintw(helpwin, line++, 2,
 		" '%c'  - text mode: keep same process info",	MBARGRAPH);
+        mvwprintw(helpwin, line++, 2,
+		" '%c'  - text mode: cgroups and related processes", MCGROUPS);
         mvwprintw(helpwin, line++, 2,
         	" '%c'  - text mode: generic info",             MPROCGEN);
         mvwprintw(helpwin, line++, 2,
@@ -2474,21 +2610,23 @@ showhelp(void)
 
 	line++;
 
-	// show context dependent help text for raw file viewing or
-	// live measurement 
+	// show context dependent help text for raw file viewing/twin mode
+	// or live measurement 
 	//
 	if (rawreadflag)
 	{
-        	mvwprintw(helpwin, line++, 2, "Raw file viewing:");
+        	mvwprintw(helpwin, line++, 2, "Raw file viewing or twin mode:");
         	mvwprintw(helpwin, line++, 2,
-			" '%c'  - show next sample in raw file", MSAMPNEXT);
+			" '%c'  - show next sample", MSAMPNEXT);
         	mvwprintw(helpwin, line++, 2,
-			" '%c'  - show previous sample in raw file", MSAMPPREV);
+			" '%c'  - show previous sample", MSAMPPREV);
         	mvwprintw(helpwin, line++, 2,
-			" '%c'  - rewind to begin of raw file", MRESET);
+			" '%c'  - rewind to begin", MRESET);
         	mvwprintw(helpwin, line++, 2,
-			" '%c'  - branch to certain time in raw file",
-								MSAMPBRANCH);
+			" '%c'  - branch to certain time", MSAMPBRANCH);
+        	mvwprintw(helpwin, line++, 2,
+        		" '%c'  - pause button to freeze or continue (twin mode)",
+								MPAUSE);
 	}
 	else
 	{
@@ -2529,7 +2667,10 @@ showhelp(void)
 	// push this keystroke back to be received by the main loop
 	//
 	if (inputkey != MQUIT)
+	{
 		ungetch(inputkey);
+		keywaiting = 1;
+	}
 
 	// remove help window
 	delwin(helpwin);
@@ -2690,3 +2831,12 @@ compnetval(const void *a, const void *b)
 
         return  0;
 }
+
+// signal catcher
+//
+static void
+getsigwinch(int signr)
+{
+        winchange = 1;
+}
+
