@@ -49,36 +49,66 @@
 #include "atop.h"
 #include "photosyst.h"
 #include "photoproc.h"
+#include "cgroups.h"
 #include "json.h"
 
 #define LEN_HP_SIZE	64
 #define LINE_BUF_SIZE	1024
 
-static void json_print_CPU(char *, struct sstat *, struct tstat *, int);
-static void json_print_cpu(char *, struct sstat *, struct tstat *, int);
-static void json_print_CPL(char *, struct sstat *, struct tstat *, int);
-static void json_print_GPU(char *, struct sstat *, struct tstat *, int);
-static void json_print_MEM(char *, struct sstat *, struct tstat *, int);
-static void json_print_SWP(char *, struct sstat *, struct tstat *, int);
-static void json_print_PAG(char *, struct sstat *, struct tstat *, int);
-static void json_print_PSI(char *, struct sstat *, struct tstat *, int);
-static void json_print_LVM(char *, struct sstat *, struct tstat *, int);
-static void json_print_MDD(char *, struct sstat *, struct tstat *, int);
-static void json_print_DSK(char *, struct sstat *, struct tstat *, int);
-static void json_print_NFM(char *, struct sstat *, struct tstat *, int);
-static void json_print_NFC(char *, struct sstat *, struct tstat *, int);
-static void json_print_NFS(char *, struct sstat *, struct tstat *, int);
-static void json_print_NET(char *, struct sstat *, struct tstat *, int);
-static void json_print_IFB(char *, struct sstat *, struct tstat *, int);
-static void json_print_NUM(char *, struct sstat *, struct tstat *, int);
-static void json_print_NUC(char *, struct sstat *, struct tstat *, int);
-static void json_print_LLC(char *, struct sstat *, struct tstat *, int);
-static void json_print_PRG(char *, struct sstat *, struct tstat *, int);
-static void json_print_PRC(char *, struct sstat *, struct tstat *, int);
-static void json_print_PRM(char *, struct sstat *, struct tstat *, int);
-static void json_print_PRD(char *, struct sstat *, struct tstat *, int);
-static void json_print_PRN(char *, struct sstat *, struct tstat *, int);
-static void json_print_PRE(char *, struct sstat *, struct tstat *, int);
+static void json_print_CPU(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_cpu(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_CPL(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_GPU(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_MEM(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_SWP(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_PAG(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_PSI(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_LVM(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_MDD(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_DSK(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_NFM(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_NFC(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_NFS(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_NET(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_IFB(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_NUM(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_NUC(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_LLC(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+
+static void json_print_CGR(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+
+static void json_print_PRG(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_PRC(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_PRM(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_PRD(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_PRN(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
+static void json_print_PRE(char *, struct sstat *, struct tstat *, int,
+                                                   struct cgchainer *, int);
 
 /*
 ** table with possible labels and the corresponding
@@ -86,37 +116,42 @@ static void json_print_PRE(char *, struct sstat *, struct tstat *, int);
 */
 struct labeldef {
 	char	*label;
-	int     valid;
-	void	(*prifunc)(char *, struct sstat *, struct tstat *, int);
+	short   valid;
+	short   cgroupref;
+	void	(*prifunc)(char *, struct sstat *,
+	                           struct tstat *, int,
+                                   struct cgchainer *, int);
 };
 
 static struct labeldef	labeldef[] = {
-	{ "CPU",	0,	json_print_CPU },
-	{ "cpu",	0,	json_print_cpu },
-	{ "CPL",	0,	json_print_CPL },
-	{ "GPU",	0,	json_print_GPU },
-	{ "MEM",	0,	json_print_MEM },
-	{ "SWP",	0,	json_print_SWP },
-	{ "PAG",	0,	json_print_PAG },
-	{ "PSI",	0,	json_print_PSI },
-	{ "LVM",	0,	json_print_LVM },
-	{ "MDD",	0,	json_print_MDD },
-	{ "DSK",	0,	json_print_DSK },
-	{ "NFM",	0,	json_print_NFM },
-	{ "NFC",	0,	json_print_NFC },
-	{ "NFS",	0,	json_print_NFS },
-	{ "NET",	0,	json_print_NET },
-	{ "IFB",	0,	json_print_IFB },
-	{ "NUM",	0,	json_print_NUM },
-	{ "NUC",	0,	json_print_NUC },
-	{ "LLC",	0,	json_print_LLC },
+	{ "CPU",	0, 0,	json_print_CPU },
+	{ "cpu",	0, 0,	json_print_cpu },
+	{ "CPL",	0, 0,	json_print_CPL },
+	{ "GPU",	0, 0,	json_print_GPU },
+	{ "MEM",	0, 0,	json_print_MEM },
+	{ "SWP",	0, 0,	json_print_SWP },
+	{ "PAG",	0, 0,	json_print_PAG },
+	{ "PSI",	0, 0,	json_print_PSI },
+	{ "LVM",	0, 0,	json_print_LVM },
+	{ "MDD",	0, 0,	json_print_MDD },
+	{ "DSK",	0, 0,	json_print_DSK },
+	{ "NFM",	0, 0,	json_print_NFM },
+	{ "NFC",	0, 0,	json_print_NFC },
+	{ "NFS",	0, 0,	json_print_NFS },
+	{ "NET",	0, 0,	json_print_NET },
+	{ "IFB",	0, 0,	json_print_IFB },
+	{ "NUM",	0, 0,	json_print_NUM },
+	{ "NUC",	0, 0,	json_print_NUC },
+	{ "LLC",	0, 0,	json_print_LLC },
 
-	{ "PRG",	0,	json_print_PRG },
-	{ "PRC",	0,	json_print_PRC },
-	{ "PRM",	0,	json_print_PRM },
-	{ "PRD",	0,	json_print_PRD },
-	{ "PRN",	0,	json_print_PRN },
-	{ "PRE",	0,	json_print_PRE },
+	{ "CGR",	0, 0,	json_print_CGR },
+
+	{ "PRG",	0, 1,	json_print_PRG },
+	{ "PRC",	0, 1,	json_print_PRC },
+	{ "PRM",	0, 1,	json_print_PRM },
+	{ "PRD",	0, 0,	json_print_PRD },
+	{ "PRN",	0, 0,	json_print_PRN },
+	{ "PRE",	0, 0,	json_print_PRE },
 };
 
 static int numlabels = sizeof labeldef / sizeof(struct labeldef);
@@ -198,9 +233,10 @@ int jsondef(char *pd)
 */
 char jsonout(time_t curtime, int numsecs,
          struct devtstat *devtstat, struct sstat *sstat,
+	 struct cgchainer *devchain, int ncgroups, int npids,
          int nexit, unsigned int noverflow, char flag)
 {
-	register int	i, j, k;
+	register int	i, j, k, cgroupref_created = 0;
 	char		header[256];
 	struct tstat	*tmp = devtstat->taskall;
 
@@ -230,11 +266,25 @@ char jsonout(time_t curtime, int numsecs,
 		if (!labeldef[i].valid)
 			continue;
 
+		/*
+		** when cgroup index is needed to map the tstat to a cgroup,
+		** once fill the tstat.gen.cgroupix variables
+		*/
+		if (supportflags & CGROUPV2 &&
+		    labeldef[i].cgroupref   && !cgroupref_created)
+		{
+			cgfillref(devtstat, devchain, ncgroups, npids);
+			cgroupref_created = 1;
+		}
+
 		/* prepare generic columns */
 		snprintf(header, sizeof header, "\"%s\"",
 			labeldef[i].label);
+
 		/* call all print-functions */
-		(labeldef[i].prifunc)(header, sstat, devtstat->taskall, devtstat->ntaskall);
+		(labeldef[i].prifunc)(header, sstat,
+				devtstat->taskall, devtstat->ntaskall,
+				devchain, ncgroups);
 	}
 
 	printf("}\n");
@@ -264,7 +314,10 @@ static void json_calc_freqscale(count_t maxfreq, count_t cnt, count_t ticks,
 	}
 }
 
-static void json_print_CPU(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_CPU(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	count_t maxfreq=0;
 	count_t cnt=0;
@@ -321,7 +374,10 @@ static void json_print_CPU(char *hp, struct sstat *ss, struct tstat *ps, int nac
 		);
 }
 
-static void json_print_cpu(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_cpu(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int i;
 	count_t maxfreq=0;
@@ -375,7 +431,10 @@ static void json_print_cpu(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_CPL(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_CPL(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	printf(", %s: {"
 		"\"lavg1\": %.2f, "
@@ -391,7 +450,10 @@ static void json_print_CPL(char *hp, struct sstat *ss, struct tstat *ps, int nac
 		ss->cpu.devint);
 }
 
-static void json_print_GPU(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_GPU(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	int	i;
 
@@ -428,7 +490,10 @@ static void json_print_GPU(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_MEM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_MEM(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	printf(", %s: {"
 		"\"physmem\": %lld, "
@@ -472,7 +537,10 @@ static void json_print_MEM(char *hp, struct sstat *ss, struct tstat *ps, int nac
 		ss->mem.anonhugepage * pagesize);
 }
 
-static void json_print_SWP(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_SWP(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	printf(", %s: {"
 		"\"totswap\": %lld, "
@@ -488,7 +556,10 @@ static void json_print_SWP(char *hp, struct sstat *ss, struct tstat *ps, int nac
 		ss->mem.commitlim * pagesize);
 }
 
-static void json_print_PAG(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_PAG(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	printf(", %s: {"
 		"\"compacts\": %lld, "
@@ -514,7 +585,10 @@ static void json_print_PAG(char *hp, struct sstat *ss, struct tstat *ps, int nac
 		ss->mem.oomkills);
 }
 
-static void json_print_PSI(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_PSI(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	if ( !(ss->psi.present) )
 		return;
@@ -554,7 +628,10 @@ static void json_print_PSI(char *hp, struct sstat *ss, struct tstat *ps, int nac
 		ss->psi.iofull.avg300, ss->psi.iofull.total);
 }
 
-static void json_print_LVM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_LVM(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int	i;
 
@@ -585,7 +662,10 @@ static void json_print_LVM(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_MDD(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_MDD(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int i;
 
@@ -616,7 +696,10 @@ static void json_print_MDD(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_DSK(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_DSK(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int	i;
 
@@ -649,7 +732,10 @@ static void json_print_DSK(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_NFM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_NFM(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int	i;
 
@@ -682,7 +768,10 @@ static void json_print_NFM(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_NFC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_NFC(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	printf(", %s: {"
 		"\"rpccnt\": %lld, "
@@ -698,7 +787,10 @@ static void json_print_NFC(char *hp, struct sstat *ss, struct tstat *ps, int nac
 		ss->nfs.client.rpcautrefresh);
 }
 
-static void json_print_NFS(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_NFS(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	printf(", %s: {"
 		"\"rpccnt\": %lld, "
@@ -734,7 +826,10 @@ static void json_print_NFS(char *hp, struct sstat *ss, struct tstat *ps, int nac
 		ss->nfs.server.rcnoca);
 }
 
-static void json_print_NET(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_NET(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int i;
 
@@ -803,7 +898,10 @@ static void json_print_NET(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_IFB(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_IFB(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int i;
 
@@ -834,7 +932,10 @@ static void json_print_IFB(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_NUM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_NUM(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int i;
 
@@ -875,7 +976,10 @@ static void json_print_NUM(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_NUC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_NUC(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int i;
 
@@ -910,7 +1014,10 @@ static void json_print_NUC(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_LLC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_LLC(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int i;
 
@@ -933,13 +1040,135 @@ static void json_print_LLC(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
+/*
+** print functions for cgroups-level statistics
+*/
+static void
+json_print_CGR(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
+{
+	if ( !(supportflags & CGROUPV2) )
+		return;
+
+	register int	i, p;
+	char		*cgrpath;
+
+        printf(", %s: [", hp);
+
+	for (i=0; i < ncgroups; i++) {
+		if (i > 0) {
+			printf(", ");
+		}
+
+                cgrpath = cggetpath(cs+i, cs, 1);
+
+                // print cgroup level metrics
+                //
+                printf( "{\"path\": \"%s\", "
+			 "\"nprocs\": %d, "
+			 "\"procsbelow\": %d, "
+			 "\"utime\": %lld, "
+			 "\"stime\": %lld, "
+			 "\"cpuweight\": %d, "
+			 "\"cpumax\": %d, "
+			 "\"cpupsisome\": %lld, "
+			 "\"cpupsitotal\": %lld, "
+
+			 "\"memcurrent\": %lld, "
+			 "\"memanon\": %lld, "
+			 "\"memfile\": %lld, "
+			 "\"memkernel\": %lld, "
+			 "\"memshmem\": %lld, "
+			 "\"memmax\": %lld, "
+			 "\"swpmax\": %lld, "
+			 "\"mempsisome\": %lld, "
+			 "\"mempsitotal\": %lld, "
+
+			 "\"diskrbytes\": %lld, "
+			 "\"diskwbytes\": %lld, "
+			 "\"diskrios\": %lld, "
+			 "\"diskwios\": %lld, "
+			 "\"diskweight\": %d, "
+			 "\"diskpsisome\": %lld, "
+			 "\"diskpsitotal\": %lld, "
+
+			 "\"pidlist\": [",
+
+			cgrpath,
+                        (cs+i)->cstat->gen.nprocs,
+                        (cs+i)->cstat->gen.procsbelow,
+                        (cs+i)->cstat->cpu.utime,
+                        (cs+i)->cstat->cpu.stime,
+                        (cs+i)->cstat->conf.cpuweight,
+                        (cs+i)->cstat->conf.cpumax,
+                        (cs+i)->cstat->cpu.somepres,
+                        (cs+i)->cstat->cpu.fullpres,
+
+                        (cs+i)->cstat->mem.current > 0 ?
+			       (cs+i)->cstat->mem.current * pagesize :
+			       (cs+i)->cstat->mem.current, 
+                        (cs+i)->cstat->mem.anon > 0 ?
+                        	(cs+i)->cstat->mem.anon * pagesize :
+                        	(cs+i)->cstat->mem.anon,
+                        (cs+i)->cstat->mem.file > 0 ?
+                        	(cs+i)->cstat->mem.file * pagesize :
+                        	(cs+i)->cstat->mem.file,
+                        (cs+i)->cstat->mem.kernel > 0 ?
+                        	(cs+i)->cstat->mem.kernel * pagesize :
+                        	(cs+i)->cstat->mem.kernel,
+                        (cs+i)->cstat->mem.shmem > 0 ?
+                        	(cs+i)->cstat->mem.shmem * pagesize :
+                        	(cs+i)->cstat->mem.shmem,
+                        (cs+i)->cstat->conf.memmax > 0 ?
+                        	(cs+i)->cstat->conf.memmax * pagesize :
+                        	(cs+i)->cstat->conf.memmax,
+                        (cs+i)->cstat->conf.swpmax > 0 ?
+                        	(cs+i)->cstat->conf.swpmax * pagesize :
+                        	(cs+i)->cstat->conf.swpmax,
+                        (cs+i)->cstat->mem.somepres,
+                        (cs+i)->cstat->mem.fullpres,
+
+                        (cs+i)->cstat->dsk.rbytes,
+                        (cs+i)->cstat->dsk.wbytes,
+                        (cs+i)->cstat->dsk.rios,
+                        (cs+i)->cstat->dsk.wios,
+                        (cs+i)->cstat->conf.dskweight,
+                        (cs+i)->cstat->dsk.somepres,
+                        (cs+i)->cstat->dsk.fullpres);
+
+		free(cgrpath);
+
+                // generate related pidlist
+                //
+                if ((cs+i)->cstat->gen.nprocs)
+                {
+                        for (p=0; p < (cs+i)->cstat->gen.nprocs; p++) {
+				if (p > 0)
+					printf(", ");
+
+                                printf("%d", (cs+i)->proclist[p]);
+			}
+
+                }
+
+                printf("]}");
+	}
+
+	printf("]");
+}
+
 /*
 ** print functions for process-level statistics
 */
-static void json_print_PRG(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_PRG(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
-	register int i, exitcode;
-	static char st[3];
+	register int	i, exitcode;
+	static char	st[3];
+	char		*cgrpath;
 
 	printf(", %s: [", hp);
 
@@ -972,9 +1201,17 @@ static void json_print_PRG(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			printf(", ");
 		}
 
-		/* using getpwuid() & getpwuid to convert ruid & euid to string seems better, but the two functions take a long time */
+		if (supportflags & CGROUPV2 && ps->gen.cgroupix != -1)
+			cgrpath = cggetpath((cs + ps->gen.cgroupix), cs, 1);
+		else
+			cgrpath = "-";
+
+		/*
+		** using getpwuid() & getpwuid to convert ruid & euid to string
+		** seems better, but the two functions take a long time
+		*/
 		printf("{\"pid\": %d, "
-			"\"name\": \"(%.19s)\", "
+			"\"cmd\": \"%.19s\", "
 			"\"state\": \"%c\", "
 			"\"ruid\": %d, "
 			"\"rgid\": %d, "
@@ -1016,15 +1253,22 @@ static void json_print_PRG(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			ps->gen.elaps,
 			!!ps->gen.isproc, /* convert to boolean */
 			ps->gen.utsname[0] ? ps->gen.utsname:"-",
-			ps->gen.cgpath[0] ? ps->gen.cgpath:"-");
+			cgrpath);
+
+		if (supportflags & CGROUPV2 && ps->gen.cgroupix != -1)
+			free(cgrpath);
 	}
 
 	printf("]");
 }
 
-static void json_print_PRC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_PRC(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
-	register int i;
+	register int	i;
+	char		*cgrpath;
 
         printf(", %s: [", hp);
 
@@ -1034,7 +1278,14 @@ static void json_print_PRC(char *hp, struct sstat *ss, struct tstat *ps, int nac
 		if (i > 0) {
 			printf(", ");
 		}
+
+		if (supportflags & CGROUPV2 && ps->gen.cgroupix != -1)
+			cgrpath = cggetpath((cs + ps->gen.cgroupix), cs, 1);
+		else
+			cgrpath = "-";
+
 		printf("{\"pid\": %d, "
+			"\"cmd\": \"%.19s\", "
 			"\"utime\": %lld, "
 			"\"stime\": %lld, "
 			"\"nice\": %d, "
@@ -1049,6 +1300,7 @@ static void json_print_PRC(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			"\"sleepavg\": %d, "
 			"\"cgroup\": \"%s\"}",
 			ps->gen.pid,
+			ps->gen.name,
 			ps->cpu.utime,
 			ps->cpu.stime,
 			ps->cpu.nice,
@@ -1061,15 +1313,22 @@ static void json_print_PRC(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			ps->cpu.nvcsw,
 			ps->cpu.nivcsw,
 			ps->cpu.sleepavg,
-			ps->gen.cgpath[0] ? ps->gen.cgpath:"-");
+			cgrpath);
+
+		if (supportflags & CGROUPV2 && ps->gen.cgroupix != -1)
+			free(cgrpath);
 	}
 
 	printf("]");
 }
 
-static void json_print_PRM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_PRM(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
-	register int i;
+	register int	i;
+	char		*cgrpath;
 
         printf(", %s: [", hp);
 
@@ -1079,7 +1338,14 @@ static void json_print_PRM(char *hp, struct sstat *ss, struct tstat *ps, int nac
 		if (i > 0) {
 			printf(", ");
 		}
+
+		if (supportflags & CGROUPV2 && ps->gen.cgroupix != -1)
+			cgrpath = cggetpath((cs + ps->gen.cgroupix), cs, 1);
+		else
+			cgrpath = "-";
+
 		printf("{\"pid\": %d, "
+			"\"cmd\": \"%.19s\", "
 			"\"vmem\": %lld, "
 			"\"rmem\": %lld, "
 			"\"vexec\": %lld, "
@@ -1095,6 +1361,7 @@ static void json_print_PRM(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			"\"pmem\": %lld, "
 			"\"cgroup\": \"%s\"}",
 			ps->gen.pid,
+			ps->gen.name,
 			ps->mem.vmem,
 			ps->mem.rmem,
 			ps->mem.vexec,
@@ -1109,13 +1376,19 @@ static void json_print_PRM(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			ps->mem.vswap,
 			ps->mem.pmem == (unsigned long long)-1LL ?
 			0:ps->mem.pmem,
-			ps->gen.cgpath[0] ? ps->gen.cgpath:"-");
+			cgrpath);
+
+		if (supportflags & CGROUPV2 && ps->gen.cgroupix != -1)
+			free(cgrpath);
 	}
 
 	printf("]");
 }
 
-static void json_print_PRD(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_PRD(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	register int i;
 
@@ -1128,12 +1401,14 @@ static void json_print_PRD(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			printf(", ");
 		}
 		printf("{\"pid\": %d, "
+			"\"cmd\": \"%.19s\", "
 			"\"rio\": %lld, "
 			"\"rsz\": %lld, "
 			"\"wio\": %lld, "
 			"\"wsz\": %lld, "
 			"\"cwsz\": %lld}",
 			ps->gen.pid,
+			ps->gen.name,
 			ps->dsk.rio, ps->dsk.rsz,
 			ps->dsk.wio, ps->dsk.wsz, ps->dsk.cwsz);
 	}
@@ -1141,7 +1416,10 @@ static void json_print_PRD(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_PRN(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_PRN(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	if (!(supportflags & NETATOP))
 		return;
@@ -1157,6 +1435,7 @@ static void json_print_PRN(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			printf(", ");
 		}
 		printf("{\"pid\": %d, "
+			"\"cmd\": \"%.19s\", "
 			"\"tcpsnd\": \"%lld\", "
 			"\"tcpssz\": \"%lld\", "
 			"\"tcprcv\": \"%lld\", "
@@ -1166,6 +1445,7 @@ static void json_print_PRN(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			"\"udprcv\": \"%lld\", "
 			"\"udprsz\": \"%lld\"}",
 			ps->gen.pid,
+			ps->gen.name,
 			ps->net.tcpsnd, ps->net.tcpssz,
 			ps->net.tcprcv, ps->net.tcprsz,
 			ps->net.udpsnd, ps->net.udpssz,
@@ -1175,7 +1455,10 @@ static void json_print_PRN(char *hp, struct sstat *ss, struct tstat *ps, int nac
 	printf("]");
 }
 
-static void json_print_PRE(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+static void
+json_print_PRE(char *hp, struct sstat *ss,
+                         struct tstat *ps, int nact,
+			 struct cgchainer *cs, int ncgroups)
 {
 	if ( !(supportflags & GPUSTAT) )
 		return;
@@ -1191,6 +1474,7 @@ static void json_print_PRE(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			printf(", ");
 		}
 		printf("{\"pid\": %d, "
+			"\"cmd\": \"%.19s\", "
 			"\"gpustate\": \"%c\", "
 			"\"nrgpus\": %d, "
 			"\"gpulist\": \"%x\", "
@@ -1200,6 +1484,7 @@ static void json_print_PRE(char *hp, struct sstat *ss, struct tstat *ps, int nac
 			"\"memcum\": %lld, "
 			"\"sample\": %lld}",
 			ps->gen.pid,
+			ps->gen.name,
 			ps->gpu.state == '\0' ? 'N':ps->gpu.state,
 			ps->gpu.nrgpus,
 			ps->gpu.gpulist,
