@@ -9,7 +9,7 @@
 ** E-mail:      gerlof.langeveld@atoptool.nl
 ** Date:        February 2007
 ** --------------------------------------------------------------------------
-** Copyright (C) 2007-2021 Gerlof Langeveld
+** Copyright (C) 2007-2024 Gerlof Langeveld
 **
 ** This program is free software; you can redistribute it and/or modify it
 ** under the terms of the GNU General Public License as published by the
@@ -37,38 +37,66 @@
 #include "atop.h"
 #include "photosyst.h"
 #include "photoproc.h"
+#include "cgroups.h"
 #include "parseable.h"
 
-void 	print_CPU(char *, struct sstat *, struct tstat *, int);
-void 	print_cpu(char *, struct sstat *, struct tstat *, int);
-void 	print_CPL(char *, struct sstat *, struct tstat *, int);
-void 	print_GPU(char *, struct sstat *, struct tstat *, int);
-void 	print_MEM(char *, struct sstat *, struct tstat *, int);
-void 	print_SWP(char *, struct sstat *, struct tstat *, int);
-void 	print_PAG(char *, struct sstat *, struct tstat *, int);
-void 	print_PSI(char *, struct sstat *, struct tstat *, int);
-void 	print_LVM(char *, struct sstat *, struct tstat *, int);
-void 	print_MDD(char *, struct sstat *, struct tstat *, int);
-void 	print_DSK(char *, struct sstat *, struct tstat *, int);
-void 	print_NFM(char *, struct sstat *, struct tstat *, int);
-void 	print_NFC(char *, struct sstat *, struct tstat *, int);
-void 	print_NFS(char *, struct sstat *, struct tstat *, int);
-void 	print_NET(char *, struct sstat *, struct tstat *, int);
-void 	print_IFB(char *, struct sstat *, struct tstat *, int);
-void 	print_NUM(char *, struct sstat *, struct tstat *, int);
-void 	print_NUC(char *, struct sstat *, struct tstat *, int);
-void 	print_LLC(char *, struct sstat *, struct tstat *, int);
+void 	print_CPU(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_cpu(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_CPL(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_GPU(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_MEM(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_SWP(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_PAG(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_PSI(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_LVM(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_MDD(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_DSK(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_NFM(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_NFC(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_NFS(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_NET(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_IFB(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_NUM(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_NUC(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_LLC(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
 
-void 	print_PRG(char *, struct sstat *, struct tstat *, int);
-void 	print_PRC(char *, struct sstat *, struct tstat *, int);
-void 	print_PRM(char *, struct sstat *, struct tstat *, int);
-void 	print_PRD(char *, struct sstat *, struct tstat *, int);
-void 	print_PRN(char *, struct sstat *, struct tstat *, int);
-void 	print_PRE(char *, struct sstat *, struct tstat *, int);
+void 	print_CGR(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+
+void 	print_PRG(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_PRC(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_PRM(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_PRD(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_PRN(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
+void 	print_PRE(char *, struct sstat *, struct tstat *, int,
+                                          struct cgchainer *, int);
 
 static void calc_freqscale(count_t, count_t, count_t, count_t *, int *);
 static char *spaceformat(char *, char *);
-static int  cgroupv2max(int, int);
 
 /*
 ** table with possible labels and the corresponding
@@ -76,37 +104,42 @@ static int  cgroupv2max(int, int);
 */
 struct labeldef {
 	char	*label;
-	int	valid;
-	void	(*prifunc)(char *, struct sstat *, struct tstat *, int);
+	short	valid;
+	short	cgroupref;
+	void	(*prifunc)(char *, struct sstat *,
+			           struct tstat *, int,
+                                   struct cgchainer *, int);
 };
 
 static struct labeldef	labeldef[] = {
-	{ "CPU",	0,	print_CPU },
-	{ "cpu",	0,	print_cpu },
-	{ "CPL",	0,	print_CPL },
-	{ "GPU",	0,	print_GPU },
-	{ "MEM",	0,	print_MEM },
-	{ "SWP",	0,	print_SWP },
-	{ "PAG",	0,	print_PAG },
-	{ "PSI",	0,	print_PSI },
-	{ "LVM",	0,	print_LVM },
-	{ "MDD",	0,	print_MDD },
-	{ "DSK",	0,	print_DSK },
-	{ "NFM",	0,	print_NFM },
-	{ "NFC",	0,	print_NFC },
-	{ "NFS",	0,	print_NFS },
-	{ "NET",	0,	print_NET },
-	{ "IFB",	0,	print_IFB },
-	{ "NUM",	0,	print_NUM },
-	{ "NUC",	0,	print_NUC },
-	{ "LLC",	0,	print_LLC },
+	{ "CPU",	0, 0,	print_CPU },
+	{ "cpu",	0, 0,	print_cpu },
+	{ "CPL",	0, 0,	print_CPL },
+	{ "GPU",	0, 0,	print_GPU },
+	{ "MEM",	0, 0,	print_MEM },
+	{ "SWP",	0, 0,	print_SWP },
+	{ "PAG",	0, 0,	print_PAG },
+	{ "PSI",	0, 0,	print_PSI },
+	{ "LVM",	0, 0,	print_LVM },
+	{ "MDD",	0, 0,	print_MDD },
+	{ "DSK",	0, 0,	print_DSK },
+	{ "NFM",	0, 0,	print_NFM },
+	{ "NFC",	0, 0,	print_NFC },
+	{ "NFS",	0, 0,	print_NFS },
+	{ "NET",	0, 0,	print_NET },
+	{ "IFB",	0, 0,	print_IFB },
+	{ "NUM",	0, 0,	print_NUM },
+	{ "NUC",	0, 0,	print_NUC },
+	{ "LLC",	0, 0,	print_LLC },
 
-	{ "PRG",	0,	print_PRG },
-	{ "PRC",	0,	print_PRC },
-	{ "PRM",	0,	print_PRM },
-	{ "PRD",	0,	print_PRD },
-	{ "PRN",	0,	print_PRN },
-	{ "PRE",	0,	print_PRE },
+	{ "CGR",	0, 0,	print_CGR },
+
+	{ "PRG",	0, 1,	print_PRG },
+	{ "PRC",	0, 1,	print_PRC },
+	{ "PRM",	0, 1,	print_PRM },
+	{ "PRD",	0, 0,	print_PRD },
+	{ "PRN",	0, 0,	print_PRN },
+	{ "PRE",	0, 0,	print_PRE },
 };
 
 static int	numlabels = sizeof labeldef/sizeof(struct labeldef);
@@ -190,9 +223,10 @@ parsedef(char *pd)
 char
 parseout(time_t curtime, int numsecs,
          struct devtstat *devtstat, struct sstat *sstat,
+         struct cgchainer *devchain, int ncgroups, int npids,
          int nexit, unsigned int noverflow, char flag)
 {
-	register int	i;
+	register int	i, cgroupref_created = 0;
 	char		datestr[32], timestr[32], header[256];
 
 	/*
@@ -208,23 +242,35 @@ parseout(time_t curtime, int numsecs,
 	{
 		if (labeldef[i].valid)
 		{
+			/*
+			** when cgroup index is needed to map the tstat to a cgroup,
+			** once fill the tstat.gen.cgroupix variables
+			*/
+			if (supportflags & CGROUPV2 &&
+			    labeldef[i].cgroupref   && !cgroupref_created)
+			{
+				cgfillref(devtstat, devchain, ncgroups, npids);
+				cgroupref_created = 1;
+			}
+
 			/*
 			** prepare generic columns
 			*/
 			convdate(curtime, datestr);
 			convtime(curtime, timestr);
 
-			snprintf(header, sizeof header, "%s %s %ld %s %s %d",
+			snprintf(header, sizeof header, "%s %s %lld %s %s %d",
 				labeldef[i].label,
 				utsname.nodename,
-				curtime,
+				(long long)curtime,
 				datestr, timestr, numsecs);
 
 			/*
-			** call a selected print-function
+			** call a selected print function
 			*/
 			(labeldef[i].prifunc)(header, sstat,
-				devtstat->taskall, devtstat->ntaskall);
+				devtstat->taskall, devtstat->ntaskall,
+				devchain, ncgroups);
 		}
 	}
 
@@ -268,7 +314,9 @@ calc_freqscale(count_t maxfreq, count_t cnt, count_t ticks,
 
 
 void
-print_CPU(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_CPU(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
         count_t maxfreq=0;
         count_t cnt=0;
@@ -313,7 +361,9 @@ print_CPU(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_cpu(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_cpu(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int i;
         count_t maxfreq=0;
@@ -351,7 +401,9 @@ print_cpu(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_CPL(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_CPL(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	printf("%s %lld %.2f %.2f %.2f %lld %lld\n",
 			hp,
@@ -364,7 +416,9 @@ print_CPL(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_GPU(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_GPU(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	int	i;
 
@@ -386,7 +440,9 @@ print_GPU(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_MEM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_MEM(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	printf(	"%s %u %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld "
    		"%lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld "
@@ -421,7 +477,9 @@ print_MEM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_SWP(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_SWP(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	printf(	"%s %u %lld %lld %lld %lld %lld %lld %lld %lld\n",
 			hp,
@@ -437,7 +495,9 @@ print_SWP(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_PAG(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_PAG(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	printf("%s %u %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld "
 	       "%lld %lld\n",
@@ -459,7 +519,9 @@ print_PAG(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_PSI(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_PSI(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	printf("%s %c %.1f %.1f %.1f %llu %.1f %.1f %.1f %llu "
 	       "%.1f %.1f %.1f %llu %.1f %.1f %.1f %llu %.1f %.1f %.1f %llu\n",
@@ -477,7 +539,9 @@ print_PSI(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_LVM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_LVM(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int	i;
 
@@ -501,7 +565,9 @@ print_LVM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_MDD(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_MDD(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int	i;
 
@@ -525,7 +591,9 @@ print_MDD(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_DSK(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_DSK(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int	i;
 
@@ -549,7 +617,9 @@ print_DSK(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_NFM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_NFM(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int	i;
 
@@ -570,7 +640,9 @@ print_NFM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_NFC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_NFC(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	printf(	"%s %lld %lld %lld %lld %lld\n",
 			hp,
@@ -582,7 +654,9 @@ print_NFC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_NFS(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_NFS(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	printf(	"%s %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld "
 	        "%lld %lld %lld %lld %lld\n",
@@ -605,7 +679,9 @@ print_NFS(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_NET(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_NET(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int 	i;
 
@@ -653,7 +729,9 @@ print_NET(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_IFB(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_IFB(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int 	i;
 
@@ -673,7 +751,9 @@ print_IFB(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_NUM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_NUM(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int 	i;
 
@@ -698,7 +778,9 @@ print_NUM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_NUC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_NUC(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int 	i;
 
@@ -720,7 +802,9 @@ print_NUC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_LLC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_LLC(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int 	i;
 
@@ -735,18 +819,113 @@ print_LLC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 	}
 }
 
+/*
+** print functions for cgroups-level statistics
+*/
+void
+print_CGR(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
+{
+	register int 	i, p;
+	char		*cgrpath;
+
+	if ( !(supportflags & CGROUPV2) )
+		return;
+
+	for (i=0; i < ncgroups; i++)
+	{
+		// print cgroup level metrics
+		//
+		cgrpath = cggetpath(devchain+i, devchain, 0);
+
+		printf(	"%s C %s %d %d %lld %lld %d %d "
+		        "%lld %lld %lld %lld %lld %lld %lld "
+		        "%lld %lld %lld %lld %d %lld %lld "
+			"%lld %lld %lld %lld\n",
+			hp, cgrpath,
+			(devchain+i)->cstat->gen.nprocs,
+			(devchain+i)->cstat->gen.procsbelow,
+			(devchain+i)->cstat->cpu.utime,
+			(devchain+i)->cstat->cpu.stime,
+			(devchain+i)->cstat->conf.cpuweight,
+			(devchain+i)->cstat->conf.cpumax,
+
+			(devchain+i)->cstat->mem.current,
+			(devchain+i)->cstat->mem.anon,
+			(devchain+i)->cstat->mem.file,
+			(devchain+i)->cstat->mem.kernel,
+			(devchain+i)->cstat->mem.shmem,
+			(devchain+i)->cstat->conf.memmax,
+			(devchain+i)->cstat->conf.swpmax,
+
+			(devchain+i)->cstat->dsk.rbytes,
+			(devchain+i)->cstat->dsk.wbytes,
+			(devchain+i)->cstat->dsk.rios,
+			(devchain+i)->cstat->dsk.wios,
+			(devchain+i)->cstat->conf.dskweight,
+			(devchain+i)->cstat->cpu.somepres,
+			(devchain+i)->cstat->cpu.fullpres,
+			(devchain+i)->cstat->mem.somepres,
+			(devchain+i)->cstat->mem.fullpres,
+			(devchain+i)->cstat->dsk.somepres,
+			(devchain+i)->cstat->dsk.fullpres);
+
+		// print related pidlist in one line
+		//
+		if ((devchain+i)->cstat->gen.nprocs)
+		{
+			printf( "%s P %s", hp, cgrpath);
+
+			for (p=0; p < (devchain+i)->cstat->gen.nprocs; p++)
+				printf(" %d", (devchain+i)->proclist[p]);
+
+			printf("\n");
+		}
+
+		free(cgrpath);
+	}
+}
+
 /*
 ** print functions for process-level statistics
 */
 void
-print_PRG(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_PRG(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
-	register int	i, exitcode;
+	register int	i, exitcode, cgrlen, cgrpathsize = 256;
 	char		namout[PNAMLEN+1+2], cmdout[CMDLEN+1+2],
-			pathout[CGRLEN+1];
+			defaultbuf[8], *cgrout=defaultbuf, *cgrpath;
+
+	// create assumed worst-case cgroup path size
+	// if it appears to be too small, it will later on be realloc'ed
+	//
+	if (supportflags & CGROUPV2)
+	{
+		cgrout = calloc(1, cgrpathsize);
+        	ptrverify(cgrout, "Malloc failed for output cgroup path\n");
+	}
 
 	for (i=0; i < nact; i++, ps++)
 	{
+		if (supportflags & CGROUPV2 && ps->gen.cgroupix != -1)	// valid cgroup index?
+		{
+			cgrpath = cggetpath((devchain + ps->gen.cgroupix), devchain, 0);
+
+			if (cgrpathsize < (cgrlen = strlen(cgrpath) + 3))
+			{
+				cgrpathsize = cgrlen;
+				cgrout      = realloc(cgrout, cgrpathsize);
+        			ptrverify(cgrout, "Realloc failed for output cgroup path\n");
+			}
+		}
+		else
+		{
+			cgrpath = "-";
+		}
+
 		if (ps->gen.excode & 0xff)      // killed by signal?
 			exitcode = (ps->gen.excode & 0x7f) + 256;
 		else
@@ -781,21 +960,41 @@ print_PRG(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 			ps->gen.ctid,
 			ps->gen.utsname[0] ? ps->gen.utsname:"-",
         		ps->gen.excode & ~(INT_MAX) ? 'N' : '-',
-			spaceformat(ps->gen.cgpath, pathout),
+			spaceformat(cgrpath, cgrout),
 			ps->gen.state == 'E' ?
 			    ps->gen.btime + ps->gen.elaps/hertz : 0,
 			ps->gen.nthridle);
+
+		if (supportflags & CGROUPV2 && ps->gen.cgroupix != -1)
+			free(cgrpath);
 	}
+
+	if (supportflags & CGROUPV2)
+		free(cgrout);
 }
 
 void
-print_PRC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_PRC(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
-	register int	i;
+	register int	i, cpumax;
 	char		namout[PNAMLEN+1+2], wchanout[20];
 
 	for (i=0; i < nact; i++, ps++)
 	{
+		if (supportflags & CGROUPV2)
+		{
+			if (ps->gen.cgroupix != -1)
+				cpumax = (devchain+ps->gen.cgroupix)->cstat->conf.cpumax;
+			else
+				cpumax = -2;
+		}
+		else
+		{
+			cpumax = -3;
+		}
+				
 		printf("%s %d %s %c %u %lld %lld %d %d %d %d %d %d %d %c "
 		       "%llu %s %llu %d %d %llu %llu\n",
 			hp,
@@ -816,21 +1015,49 @@ print_PRC(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 			ps->cpu.rundelay,
 			spaceformat(ps->cpu.wchan, wchanout),
 			ps->cpu.blkdelay,
-			cgroupv2max(ps->gen.isproc, ps->cpu.cgcpumax),
-			cgroupv2max(ps->gen.isproc, ps->cpu.cgcpumaxr),
+			cpumax,
+			-2,	// most restrictive cpumax no longer supported
 			ps->cpu.nvcsw,
 			ps->cpu.nivcsw);
 	}
 }
 
 void
-print_PRM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_PRM(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
-	register int 	i;
+	register int 	i, memmax, swpmax;
 	char		namout[PNAMLEN+1+2];
 
 	for (i=0; i < nact; i++, ps++)
 	{
+		if (supportflags & CGROUPV2)
+		{
+			if (ps->gen.cgroupix != -1)
+			{
+				memmax = (devchain+ps->gen.cgroupix)->cstat->conf.memmax;
+
+				if (memmax > 0)
+					memmax *= pagesize / 1024;
+
+				swpmax = (devchain+ps->gen.cgroupix)->cstat->conf.swpmax;
+
+				if (swpmax > 0)
+					swpmax *= pagesize / 1024;
+			}
+			else
+			{
+				memmax = -2;
+				swpmax = -2;
+			}
+		}
+		else
+		{
+			memmax = -3;
+			swpmax = -3;
+		}
+
 		printf("%s %d %s %c %u %lld %lld %lld %lld %lld %lld "
 		       "%lld %lld %lld %lld %lld %d %c %lld %lld %d %d %d %d\n",
 			hp,
@@ -852,17 +1079,17 @@ print_PRM(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 			ps->gen.tgid,
 			ps->gen.isproc ? 'y':'n',
 			ps->mem.pmem == (unsigned long long)-1LL ?
-							0:ps->mem.pmem,
+						0:ps->mem.pmem,
 			ps->mem.vlock,
-			cgroupv2max(ps->gen.isproc, ps->mem.cgmemmax),
-			cgroupv2max(ps->gen.isproc, ps->mem.cgmemmaxr),
-			cgroupv2max(ps->gen.isproc, ps->mem.cgswpmax),
-			cgroupv2max(ps->gen.isproc, ps->mem.cgswpmaxr));
+			memmax, -2,
+			swpmax, -2);
 	}
 }
 
 void
-print_PRD(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_PRD(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int	i;
 	char		namout[PNAMLEN+1+2];
@@ -884,7 +1111,9 @@ print_PRD(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_PRN(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_PRN(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int	i;
 	char		namout[PNAMLEN+1+2];
@@ -908,7 +1137,9 @@ print_PRN(char *hp, struct sstat *ss, struct tstat *ps, int nact)
 }
 
 void
-print_PRE(char *hp, struct sstat *ss, struct tstat *ps, int nact)
+print_PRE(char *hp, struct sstat *ss,
+                    struct tstat *ps, int nact,
+                    struct cgchainer *devchain, int ncgroups)
 {
 	register int	i;
 	char		namout[PNAMLEN+1+2];
@@ -982,18 +1213,3 @@ spaceformat(char *istr, char *ostr)
 
 	return ostr;
 }
-
-/*
-** return proper integer for cgroup v2 maximum values
-*/
-static int
-cgroupv2max(int isproc, int max)
-{
-	if (! (supportflags&CGROUPV2))
-		return -3;
-
-	if (! isproc)
-		return -2;
-
-	return max;
-}
