@@ -45,6 +45,7 @@
 #include <dirent.h>
 #include <sys/ioctl.h>
 #include <sys/sysmacros.h>
+#include <sys/resource.h>
 #include <limits.h>
 
 #define SCALINGMAXCPU	8	// threshold for scaling info per CPU
@@ -69,12 +70,6 @@
 
 #define	MAXCNT	64
 
-/* return value of isdisk() */
-#define	NONTYPE	0
-#define	DSKTYPE	1
-#define	MDDTYPE	2
-#define	LVMTYPE	3
-
 /* recognize numa node */
 #define	NUMADIR	"/sys/devices/system/node"
 
@@ -104,7 +99,7 @@ static void	getperfevents(struct cpustat *);
 static int	get_infiniband(struct ifbstat *);
 static int	get_ksm(struct sstat *);
 
-static int	isdisk(unsigned int, unsigned int,
+static int	isdisk_name(unsigned int, unsigned int,
 			char *, struct perdsk *, int);
 
 static struct bitmask *numa_allocate_cpumask(void);
@@ -293,7 +288,7 @@ photosyst(struct sstat *si)
 	FILE 		*fp;
 	DIR		*dirp;
 	struct dirent	*dentry;
-	char		linebuf[1024], nam[64], origdir[1024];
+	char		linebuf[1024], nam[64], origdir[4096];
 	unsigned int	major, minor;
 	struct shm_info	shminfo;
 #if	HTTPSTATS
@@ -450,86 +445,94 @@ photosyst(struct sstat *si)
         static char fn[512];
         int didone=0;
 
-	if (si->cpu.nrcpu <= SCALINGMAXCPU)
-	{
-            for (i = 0; i < si->cpu.nrcpu; ++i)
-            {
-                long long f=0;
+	// check governor statistics
+        for (i = 0; i < si->cpu.nrcpu; ++i)
+        {
+		long long f=0;
 
-                snprintf(fn, sizeof fn,
+		snprintf(fn, sizeof fn,
                    "/sys/devices/system/cpu/cpu%d/cpufreq/stats/time_in_state",
                    i);
 
-                if ((fp=fopen(fn, "r")) != 0)
-                {
-                        long long hits=0;
-                        long long maxfreq=0;
-                        long long cnt=0;
-                        long long sum=0;
+		if ((fp=fopen(fn, "r")) != 0)
+		{
+                    long long hits=0;
+                    long long maxfreq=0;
+                    long long cnt=0;
+                    long long sum=0;
 
-                        while (fscanf(fp, "%lld %lld", &f, &cnt) == 2)
-                        {
-                                f	/= 1000;
-                                sum 	+= (f*cnt);
-                                hits	+= cnt;
+                    while (fscanf(fp, "%lld %lld", &f, &cnt) == 2)
+                    {
+                            f		/= 1000;
+                            sum 	+= (f*cnt);
+                            hits	+= cnt;
 
-                                if (f > maxfreq)
-                        		maxfreq=f;
-                        	didone=1;
-                        }
+                            if (f > maxfreq)
+                    		maxfreq=f;
+                    }
 
-	                si->cpu.cpu[i].freqcnt.maxfreq	= maxfreq;
-	                si->cpu.cpu[i].freqcnt.cnt	= sum;
-	                si->cpu.cpu[i].freqcnt.ticks	= hits;
+                    si->cpu.cpu[i].freqcnt.maxfreq = maxfreq;
+                    si->cpu.cpu[i].freqcnt.cnt	   = sum;
+                    si->cpu.cpu[i].freqcnt.ticks   = hits;
 
-                        fclose(fp);
-                }
+                    fclose(fp);
+	            didone=1;
+		}
 		else
-		{    // governor statistics not available
-                     snprintf(fn, sizeof fn, 
-                      "/sys/devices/system/cpu/cpu%d/cpufreq/cpuinfo_max_freq",
-		      i);
-
-                        if ((fp=fopen(fn, "r")) != 0)
-                        {
-                                if (fscanf(fp, "%lld", &f) == 1)
-                                {
-  					// convert KHz to MHz
-	                                si->cpu.cpu[i].freqcnt.maxfreq =f/1000; 
-                                }
-
-                                fclose(fp);
-                        }
-                        else 
-                        {
-	                        si->cpu.cpu[i].freqcnt.maxfreq=0;
-                        }
-
-                       snprintf(fn, sizeof fn,
-                       "/sys/devices/system/cpu/cpu%d/cpufreq/scaling_cur_freq",
-		       i);
-        
-                        if ((fp=fopen(fn, "r")) != 0)
-                        {
-                                if (fscanf(fp, "%lld", &f) == 1)
-                                {
-   					// convert KHz to MHz
-                                        si->cpu.cpu[i].freqcnt.cnt   = f/1000;
-                                        si->cpu.cpu[i].freqcnt.ticks = 0;
-                                        didone=1;
-                                }
-
-                                fclose(fp);
-                        }
-                        else
-                        {
-                                si->cpu.cpu[i].freqcnt.cnt	= 0;
-                                si->cpu.cpu[i].freqcnt.ticks 	= 0;
-                        }
-                }
-            } // for all CPUs
+		{
+			break;
+		}
 	}
 
+        if (!didone)     // did not get processor freq statistics yet
+	{
+		long long f=0;
+
+        	for (i = 0; i < si->cpu.nrcpu; ++i)
+        	{
+        		snprintf(fn, sizeof fn, 
+               		       "/sys/devices/system/cpu/cpu%d/cpufreq/cpuinfo_max_freq",
+			      i);
+
+               	 	if ((fp=fopen(fn, "r")) != 0)
+                	{
+                        	if (fscanf(fp, "%lld", &f) == 1)
+                        	{
+  					// convert KHz to MHz
+	                        	si->cpu.cpu[i].freqcnt.maxfreq =f/1000; 
+                        	}
+
+                        	fclose(fp);
+                	}
+                	else 
+                	{
+	                	si->cpu.cpu[i].freqcnt.maxfreq=0;
+                	}
+
+                	snprintf(fn, sizeof fn,
+                       		"/sys/devices/system/cpu/cpu%d/cpufreq/scaling_cur_freq",
+		       	i);
+        
+                	if ((fp=fopen(fn, "r")) != 0)
+                	{
+                 		if (fscanf(fp, "%lld", &f) == 1)
+                        	{
+   					// convert KHz to MHz
+                                	si->cpu.cpu[i].freqcnt.cnt   = f/1000;
+                                	si->cpu.cpu[i].freqcnt.ticks = 0;
+                        	}
+	
+                        	fclose(fp);
+                        	didone=1;
+                	}
+                	else
+                	{
+                        	si->cpu.cpu[i].freqcnt.cnt	= 0;
+                        	si->cpu.cpu[i].freqcnt.ticks 	= 0;
+                	}
+		}
+        } // for all CPUs
+
         if (!didone)     // did not get processor freq statistics.
                          // use /proc/cpuinfo
         {
@@ -1443,7 +1446,7 @@ photosyst(struct sstat *si)
 			*/
 			if (nr == 8)	/* full stats-line ? */
 			{
-				if ( isdisk(0, 0, diskname,
+				if ( isdisk_name(0, 0, diskname,
 				                 &(si->dsk.dsk[i]),
 						 MAXDKNAM) != DSKTYPE)
 				       continue;
@@ -1506,7 +1509,7 @@ photosyst(struct sstat *si)
 				** or just one of the partitions of a disk (to be
 				** skipped)
 				*/
-				switch ( isdisk(major, minor, diskname,
+				switch ( isdisk_name(major, minor, diskname,
 							 &tmpdsk, MAXDKNAM) )
 				{
 				   case NONTYPE:
@@ -2214,7 +2217,7 @@ lvmmapname(unsigned int major, unsigned int minor,
 }
 
 /*
-** this table is used in the function isdisk()
+** this table is used in the functions isdisk_name() and isdick_major()
 **
 ** table contains the names (in regexp format) of disks
 ** to be recognized, together with a function to modify
@@ -2227,32 +2230,33 @@ static struct {
 	regex_t	compreg;
 	void	(*modname)(unsigned int, unsigned int,
 				char *, struct perdsk *, int);
+	int	major;	// to be filled via isdisk_name()
 	int	retval;
 } validdisk[] = {
-	{ "^ram[0-9][0-9]*$",			{0},  (void *)0,   NONTYPE, },
-	{ "^loop[0-9][0-9]*$",			{0},  (void *)0,   NONTYPE, },
-	{ "^sd[a-z][a-z]*$",			{0},  nullmodname, DSKTYPE, },
-	{ "^dm-[0-9][0-9]*$",			{0},  lvmmapname,  LVMTYPE, },
-	{ "^md[0-9][0-9]*$",			{0},  nullmodname, MDDTYPE, },
-	{ "^vd[a-z][a-z]*$",                    {0},  nullmodname, DSKTYPE, },
-	{ "^nvme[0-9][0-9]*n[0-9][0-9]*$",	{0},  nullmodname, DSKTYPE, },
-	{ "^nvme[0-9][0-9]*c[0-9][0-9]*n[0-9][0-9]*$", {0}, nullmodname, DSKTYPE, },
-	{ "^nbd[0-9][0-9]*$",			{0},  nullmodname, DSKTYPE, },
-	{ "^hd[a-z]$",				{0},  nullmodname, DSKTYPE, },
-	{ "^rd/c[0-9][0-9]*d[0-9][0-9]*$",	{0},  nullmodname, DSKTYPE, },
-	{ "^cciss/c[0-9][0-9]*d[0-9][0-9]*$",	{0},  nullmodname, DSKTYPE, },
-	{ "^fio[a-z][a-z]*$",			{0},  nullmodname, DSKTYPE, },
-	{ "/host.*/bus.*/target.*/lun.*/disc",	{0},  abbrevname1, DSKTYPE, },
-	{ "^xvd[a-z][a-z]*[0-9]*$",		{0},  nullmodname, DSKTYPE, },
-	{ "^dasd[a-z][a-z]*$",			{0},  nullmodname, DSKTYPE, },
-	{ "^mmcblk[0-9][0-9]*$",		{0},  nullmodname, DSKTYPE, },
-	{ "^emcpower[a-z][a-z]*$",		{0},  nullmodname, DSKTYPE, },
-	{ "^rbd[0-9][0-9]*$",			{0},  nullmodname, DSKTYPE, },
-	{ "^rbd[0-9][0-9]*p[0-9][0-9]*$",	{0},  nullmodname, DSKTYPE, },
+	{ "^ram[0-9][0-9]*$",			{0},  (void *)0,   0, NONTYPE, },
+	{ "^loop[0-9][0-9]*$",			{0},  (void *)0,   0, NONTYPE, },
+	{ "^sd[a-z][a-z]*$",			{0},  nullmodname, 0, DSKTYPE, },
+	{ "^dm-[0-9][0-9]*$",			{0},  lvmmapname,  0, LVMTYPE, },
+	{ "^md[0-9][0-9]*$",			{0},  nullmodname, 0, MDDTYPE, },
+	{ "^vd[a-z][a-z]*$",                    {0},  nullmodname, 0, DSKTYPE, },
+	{ "^nvme[0-9][0-9]*n[0-9][0-9]*$",	{0},  nullmodname, 0, DSKTYPE, },
+	{ "^nvme[0-9][0-9]*c[0-9][0-9]*n[0-9][0-9]*$", {0}, nullmodname, 0, DSKTYPE, },
+	{ "^nbd[0-9][0-9]*$",			{0},  nullmodname, 0, DSKTYPE, },
+	{ "^hd[a-z]$",				{0},  nullmodname, 0, DSKTYPE, },
+	{ "^rd/c[0-9][0-9]*d[0-9][0-9]*$",	{0},  nullmodname, 0, DSKTYPE, },
+	{ "^cciss/c[0-9][0-9]*d[0-9][0-9]*$",	{0},  nullmodname, 0, DSKTYPE, },
+	{ "^fio[a-z][a-z]*$",			{0},  nullmodname, 0, DSKTYPE, },
+	{ "/host.*/bus.*/target.*/lun.*/disc",	{0},  abbrevname1, 0, DSKTYPE, },
+	{ "^xvd[a-z][a-z]*[0-9]*$",		{0},  nullmodname, 0, DSKTYPE, },
+	{ "^dasd[a-z][a-z]*$",			{0},  nullmodname, 0, DSKTYPE, },
+	{ "^mmcblk[0-9][0-9]*$",		{0},  nullmodname, 0, DSKTYPE, },
+	{ "^emcpower[a-z][a-z]*$",		{0},  nullmodname, 0, DSKTYPE, },
+	{ "^rbd[0-9][0-9]*$",			{0},  nullmodname, 0, DSKTYPE, },
+	{ "^rbd[0-9][0-9]*p[0-9][0-9]*$",	{0},  nullmodname, 0, DSKTYPE, },
 };
 
 static int
-isdisk(unsigned int major, unsigned int minor,
+isdisk_name(unsigned int major, unsigned int minor,
            char *curname, struct perdsk *px, int maxlen)
 {
 	static int	firstcall = 1;
@@ -2273,6 +2277,12 @@ isdisk(unsigned int major, unsigned int minor,
 	{
 		if (regexec(&validdisk[i].compreg, curname, 0, NULL, 0) == 0)
 		{
+			/*
+			** store major number of driver
+			*/
+			if (major)
+				validdisk[i].major = major;
+
 			/*
 			** name-string recognized; modify name-string
 			*/
@@ -2288,6 +2298,24 @@ isdisk(unsigned int major, unsigned int minor,
 }
 
 
+int
+isdisk_major(unsigned int major)
+{
+	register int	i;
+
+	/*
+	** search for the major number of the disk
+	*/
+	for (i=0; i < sizeof validdisk/sizeof validdisk[0]; i++)
+	{
+		if (major == validdisk[i].major)
+			return validdisk[i].retval;
+	}
+
+	return NONTYPE;
+}
+
+
 /*
 ** get stats of all InfiniBand ports below
 **    /sys/class/infiniband/<controller>/ports/<port#>/....
@@ -2783,10 +2811,24 @@ getperfevents(struct cpustat *cs)
 	if (firstcall)
 	{
 		struct perf_event_attr  pea;
-		int			success = 0;
+		int			success=0, minfds = cs->nrcpu*2 + 32;
+		struct rlimit		rlim;
 
 		firstcall = 0;
 
+		/*
+		** for perf events two file descriptors will
+		** be opened permanently per CPU, so take care
+		** that enough open files are allowed for this process
+		*/
+		getrlimit(RLIMIT_NOFILE, &rlim);
+
+		if (rlim.rlim_cur < minfds)	// default not enough?
+		{
+			rlim.rlim_cur = minfds;
+			(void) setrlimit(RLIMIT_NOFILE, &rlim);
+		}
+
 		/*
 		** allocate space for per-cpu file descriptors
 		*/
@@ -2824,7 +2866,6 @@ getperfevents(struct cpustat *cs)
 		if (! droprootprivs())
 			mcleanstop(42, "failed to drop root privs\n");
 
-
 		/*
 		** all failed (probably no kernel support)?
 		*/
@@ -2860,7 +2901,7 @@ getperfevents(struct cpustat *cs)
 		if (*(fdi+i) != -1)
 		{
                 	liResult = read(*(fdi+i), &(cs->cpu[i].instr), sizeof(count_t));
-                        cs->all.instr += cs->cpu[i].instr;
+
 			if(liResult < 0)
 			{
 				char lcMessage[64];
@@ -2870,9 +2911,13 @@ getperfevents(struct cpustat *cs)
 				           __FILE__, __LINE__, errno);
 				fprintf(stderr, "%s", lcMessage);
 			}
+			else
+			{
+                        	cs->all.instr += cs->cpu[i].instr;
+			}
 
                 	liResult = read(*(fdc+i), &(cs->cpu[i].cycle), sizeof(count_t));
-                        cs->all.cycle += cs->cpu[i].cycle;
+
 			if(liResult < 0)
 			{
 				char lcMessage[64];
@@ -2882,6 +2927,10 @@ getperfevents(struct cpustat *cs)
 				           __FILE__, __LINE__, errno );
 				fprintf(stderr, "%s", lcMessage);
 			}
+			else
+			{
+                        	cs->all.cycle += cs->cpu[i].cycle;
+			}
 		}
         }
 }
