@@ -7,8 +7,9 @@
 ** Author:      Gerlof Langeveld
 ** E-mail:      gerlof.langeveld@atoptool.nl
 ** Date:        September 2002
+** Date:        February 2024 (added cgroup support)
 ** --------------------------------------------------------------------------
-** Copyright (C) 2000-2010 Gerlof Langeveld
+** Copyright (C) 2000-2024 Gerlof Langeveld
 **
 ** This program is free software; you can redistribute it and/or modify it
 ** under the terms of the GNU General Public License as published by the
@@ -49,9 +50,9 @@
 
 #include "atop.h"
 #include "photoproc.h"
-#include "showgeneric.h"
-#include "photoproc.h"
 #include "photosyst.h"
+#include "cgroups.h"
+#include "showgeneric.h"
 #include "rawlog.h"
 
 #define	BASEPATH	"/var/log/atop"  
@@ -60,6 +61,10 @@
 static int	getrawrec  (int, struct rawrecord *, int);
 static int	getrawsstat(int, struct sstat *, int);
 static int	getrawtstat(int, struct tstat *, int, int);
+static int	getrawcstat(int, struct cgchainer **,
+			unsigned long, unsigned long,
+                        unsigned long, int, int);
+
 static int	rawwopen(void);
 static int	readchunk(int, void *, int);
 static int	lookslikedatetome(char *);
@@ -73,6 +78,7 @@ static void	try_other_version(int, int);
 char
 rawwrite(time_t curtime, int numsecs, 
          struct devtstat *devtstat, struct sstat *sstat,
+	 struct cgchainer *devchain, int ncgroups, int npids,
          int nexit, unsigned int noverflow, char flag)
 {
 	static int		rawfd = -1;
@@ -80,11 +86,16 @@ rawwrite(time_t curtime, int numsecs,
 	int			rv;
 	struct stat		filestat;
 
-	Byte			scompbuf[sizeof(struct sstat)], *pcompbuf;
-	unsigned long		scomplen = sizeof scompbuf;
-	unsigned long		pcomplen = sizeof(struct tstat) *
-						devtstat->ntaskall;
-	struct iovec 		iov[3];
+	Byte			scompbuf[sizeof(struct sstat)], *pcompbuf,
+				*ccompbuf = NULL, *icompbuf = NULL;
+
+	unsigned long		soriglen = sizeof scompbuf, scomplen = compressBound(soriglen),
+				poriglen, pcomplen,
+				coriglen, ccomplen,
+				ioriglen, icomplen;
+
+	struct iovec 		iov[5];
+	int			nrvectors;
 
 	/*
 	** first call:
@@ -101,21 +112,71 @@ rawwrite(time_t curtime, int numsecs,
 	(void) fstat(rawfd, &filestat);
 
 	/*
-	** compress system- and process-level statistics
+	** compress system level metrics
 	*/
-	rv = compress(scompbuf, &scomplen,
-				(Byte *)sstat, (unsigned long)sizeof *sstat);
+	rv = compress(scompbuf, &scomplen, (Byte *)sstat, soriglen);
 
-	testcompval(rv, "compress");
+	testcompval(rv, "compress system stats");
+
+	/*
+	** compress process level metrics
+	*/
+	poriglen = sizeof(struct tstat) * devtstat->ntaskall;
+	pcomplen = compressBound(poriglen);
 
 	pcompbuf = malloc(pcomplen);
 
-	ptrverify(pcompbuf, "Malloc failed for compression buffer\n");
+	ptrverify(pcompbuf, "Malloc failed for process compression buffer\n");
 
-	rv = compress(pcompbuf, &pcomplen, (Byte *)devtstat->taskall,
-						(unsigned long)pcomplen);
+	rv = compress(pcompbuf, &pcomplen, (Byte *)devtstat->taskall, poriglen);
 
-	testcompval(rv, "compress");
+	testcompval(rv, "compress processes");
+
+	/*
+	** compress cgroup level metrics
+	*/
+	if (supportflags & CGROUPV2)
+	{
+		/*
+		** calculate the size of all contiguous cstat structs
+		** and compress
+		*/
+		coriglen = (char *)(devchain+ncgroups-1)->cstat -
+			   (char *) devchain->cstat +
+			           (devchain+ncgroups-1)->cstat->gen.structlen;
+
+		ccomplen  = compressBound(coriglen);
+
+		ccompbuf = malloc(ccomplen);
+
+		ptrverify(ccompbuf, "Malloc failed for cgroup compression buffer\n");
+
+		rv = compress(ccompbuf, &ccomplen, (Byte *)devchain->cstat, coriglen);
+
+		testcompval(rv, "compress cgroups");
+
+		/*
+		** calculate the size of the cgroups pidlist
+		** and compress
+		*/
+		ioriglen = npids * sizeof(pid_t);
+		icomplen = compressBound(ioriglen);
+
+		icompbuf = malloc(icomplen);
+
+		ptrverify(icompbuf, "Malloc failed for cgroup compression pidlist\n");
+
+		rv = compress(icompbuf, &icomplen, (Byte *)devchain->proclist, ioriglen);
+
+		nrvectors = 5;
+	}
+	else
+	{
+		coriglen  = ccomplen = 0;
+		ioriglen  = icomplen = 0;
+	
+		nrvectors = 3;
+	}
 
 	/*
 	** fill record header and write to file
@@ -136,8 +197,13 @@ rawwrite(time_t curtime, int numsecs,
 	rr.totslpu	= devtstat->totslpu;
 	rr.totidle	= devtstat->totidle;
 	rr.totzomb	= devtstat->totzombie;
+	rr.ncgroups	= ncgroups;
+	rr.ncgpids	= npids;
 	rr.scomplen	= scomplen;
 	rr.pcomplen	= pcomplen;
+	rr.ccomplen	= ccomplen;
+	rr.coriglen	= coriglen;
+	rr.icomplen	= icomplen;
 
 	if (flag&RRBOOT)
 		rr.flags |= RRBOOT;
@@ -164,11 +230,11 @@ rawwrite(time_t curtime, int numsecs,
 		rr.flags |= RRGPUSTAT;
 
 	/*
-	** use 1-writev to make record operation atomic
+	** use writev to make record operation atomic
 	** to avoid write uncompleted record data.
 	*/
 	iov[0].iov_base = &rr;
-	iov[0].iov_len  = sizeof (rr);
+	iov[0].iov_len  = sizeof(rr);
 
 	iov[1].iov_base = scompbuf;
 	iov[1].iov_len  = scomplen;
@@ -176,7 +242,13 @@ rawwrite(time_t curtime, int numsecs,
 	iov[2].iov_base = pcompbuf;
 	iov[2].iov_len  = pcomplen;
 
-	if ( writev(rawfd, iov, 3) == -1)
+	iov[3].iov_base = ccompbuf;
+	iov[3].iov_len  = ccomplen;
+
+	iov[4].iov_base = icompbuf;
+	iov[4].iov_len  = icomplen;
+
+	if ( writev(rawfd, iov, nrvectors) == -1)
 	{
 		fprintf(stderr, "%s - ", rawname);
 		if ( ftruncate(rawfd, filestat.st_size) == -1)
@@ -191,6 +263,12 @@ rawwrite(time_t curtime, int numsecs,
 
 	free(pcompbuf);
 
+	if (supportflags & CGROUPV2)
+	{
+		free(ccompbuf);
+		free(icompbuf);
+	}
+
 	return '\0';
 }
 
@@ -213,67 +291,85 @@ rawwopen()
 {
 	struct rawheader	rh;
 	int			fd;
+	struct stat		filestats;
 
 	/*
 	** check if the file exists already
 	*/
-	if ( (fd = open(rawname, O_RDWR)) >= 0)
+	if ( (fd = open(rawname, O_RDWR)) >= 0 )
 	{
 		/*
-		** read and verify header record
+		** check if the file already contains a file header (and records)
 		*/
-		if ( read(fd, &rh, sizeof rh) < sizeof rh)
-			mcleanstop(7, "%s - cannot read header\n", rawname);
-
-		if (rh.magic != MYMAGIC)
-			mcleanstop(7, "file %s exists but does not contain raw "
-				"atop output (wrong magic number)\n", rawname);
-
-		if ( rh.sstatlen	!= sizeof(struct sstat)		||
-		     rh.tstatlen	!= sizeof(struct tstat)		||
-	    	     rh.rawheadlen	!= sizeof(struct rawheader)	||
-		     rh.rawreclen	!= sizeof(struct rawrecord)	  )
+		if (fstat(fd, &filestats) == 0 && filestats.st_size > 0)
 		{
-			fprintf(stderr,
-				"existing file %s has incompatible header\n",
-				rawname);
+			/*
+			** read and verify raw file header
+			*/
+			if ( read(fd, &rh, sizeof rh) < sizeof rh)
+				mcleanstop(7, "%s - cannot read header\n", rawname);
 
-			if (rh.aversion & 0x8000 &&
-			   (rh.aversion & 0x7fff) != getnumvers())
+			if (rh.magic != MYMAGIC)
+				mcleanstop(7, "file %s exists but does not contain raw "
+					"atop output (wrong magic number)\n", rawname);
+
+			if ( rh.sstatlen	!= sizeof(struct sstat)		||
+			     rh.tstatlen	!= sizeof(struct tstat)		||
+			     rh.cstatlen	!= sizeof(struct cstat)		||
+		    	     rh.rawheadlen	!= sizeof(struct rawheader)	||
+			     rh.rawreclen	!= sizeof(struct rawrecord)	  )
 			{
 				fprintf(stderr,
-					"(created by version %d.%d - "
-					"current version %d.%d)\n",
-					(rh.aversion >> 8) & 0x7f,
-					 rh.aversion & 0xff,
-					 getnumvers() >> 8,
-					 getnumvers() & 0x7f);
+					"existing file %s has incompatible header\n",
+					rawname);
+
+				if (rh.aversion & 0x8000 &&
+				   (rh.aversion & 0x7fff) != getnumvers())
+				{
+					fprintf(stderr,
+						"(created by version %d.%d - "
+						"current version %d.%d)\n",
+						(rh.aversion >> 8) & 0x7f,
+						 rh.aversion & 0xff,
+						 getnumvers() >> 8,
+						 getnumvers() & 0x7f);
+				}
+
+				cleanstop(7);
 			}
 
+			/*
+			** jump to end of file, being prepared to extend with more records
+			*/
+			(void) lseek(fd, (off_t) 0, SEEK_END);
+
+			return fd;
+		}
+	}
+	else
+	{
+		/*
+		** file does not exist (or can not be opened)
+		*/
+		if ( (fd = creat(rawname, 0666)) == -1)
+		{
+			fprintf(stderr, "%s - ", rawname);
+			perror("create raw file");
 			cleanstop(7);
 		}
-
-		(void) lseek(fd, (off_t) 0, SEEK_END);
-
-		return fd;
 	}
 
 	/*
-	** file does not exist (or can not be opened)
+	** empty file is opened now
+	** write a raw file header
 	*/
-	if ( (fd = creat(rawname, 0666)) == -1)
-	{
-		fprintf(stderr, "%s - ", rawname);
-		perror("create raw file");
-		cleanstop(7);
-	}
-
 	memset(&rh, 0, sizeof rh);
 
 	rh.magic	= MYMAGIC;
 	rh.aversion	= getnumvers() | 0x8000;
 	rh.sstatlen	= sizeof(struct sstat);
 	rh.tstatlen	= sizeof(struct tstat);
+	rh.cstatlen	= sizeof(struct cstat);
 	rh.rawheadlen	= sizeof(struct rawheader);
 	rh.rawreclen	= sizeof(struct rawrecord);
 	rh.supportflags	= supportflags | RAWLOGNG;
@@ -304,12 +400,14 @@ rawwopen()
 int
 rawread(void)
 {
+	static struct devtstat	devtstat;
+
 	int			i, j, rawfd, len, isregular = 1;
 	char			*py;
 	struct rawheader	rh;
 	struct rawrecord	rr;
 	struct sstat		sstat;
-	static struct devtstat	devtstat;
+	struct cgchainer	*devchain = NULL;
 
 	struct stat		filestat;
 
@@ -486,11 +584,17 @@ rawread(void)
 	*/
 	if (rh.sstatlen   != sizeof(struct sstat)		||
 	    rh.tstatlen   != sizeof(struct tstat)		||
+	    rh.cstatlen   != sizeof(struct cstat)		||
 	    rh.rawheadlen != sizeof(struct rawheader)		||
 	    rh.rawreclen  != sizeof(struct rawrecord)		  )
 	{
+		fprintf(stderr, "sstatlen: %d/%lu\n", rh.sstatlen, sizeof(struct sstat));
+		fprintf(stderr, "cstatlen: %d/%lu\n", rh.cstatlen, sizeof(struct cstat));
+		fprintf(stderr, "tstatlen: %d/%lu\n", rh.tstatlen, sizeof(struct tstat));
+		fprintf(stderr, "headlen:  %d/%lu\n", rh.rawheadlen, sizeof(struct rawheader));
+		fprintf(stderr, "reclen:   %d/%lu\n", rh.rawreclen, sizeof(struct rawrecord));
 		fprintf(stderr,
-			"raw file %s has incompatible format\n", rawname);
+			"\nraw file %s has incompatible format\n", rawname);
 
 		if (rh.aversion & 0x8000 &&
        		   (rh.aversion & 0x7fff) != getnumvers())
@@ -577,8 +681,7 @@ rawread(void)
 			** (only happens for the first record)
 			*/
 			if (begintime <= SECONDSINDAY)
-				begintime = normalize_epoch(cursortime,
-								begintime);
+				begintime = normalize_epoch(cursortime, begintime);
 
 			if (endtime && endtime <= SECONDSINDAY)
 				endtime = normalize_epoch(cursortime, endtime);
@@ -589,8 +692,7 @@ rawread(void)
 			*/
 			if (isregular)
 			{
-				*(offlist+offcur) = lseek(rawfd, 0, SEEK_CUR) -
-								rh.rawreclen;
+				*(offlist+offcur) = lseek(rawfd, 0, SEEK_CUR) - rh.rawreclen;
 
 				if ( ++offcur >= offsize )
 				{
@@ -618,14 +720,19 @@ rawread(void)
 					off_t next_pos;
 
 					lastcmd = 1;
-					next_pos = lseek(rawfd, rr.scomplen+rr.pcomplen, SEEK_CUR);
+					next_pos = lseek(rawfd, rr.scomplen+rr.pcomplen+rr.ccomplen+rr.icomplen, SEEK_CUR);
+
 					if ((curr_pos >> READAHEADOFF) != (next_pos >> READAHEADOFF))
 					{
 						int liResult;
+
 						/* just read READAHEADSIZE bytes into page cache */
 						char *buf = malloc(READAHEADSIZE);
+
 						ptrverify(buf, "Malloc failed for readahead");
+
 						liResult = pread(rawfd, buf, READAHEADSIZE, next_pos & ~(READAHEADSIZE - 1));
+
 						if(liResult == -1)
 						{
 							char lcMessage[64];
@@ -635,6 +742,7 @@ rawread(void)
 							          __FILE__, __LINE__, errno);
 							fprintf(stderr, "%s", lcMessage);
 						}
+
 						free(buf);
 					}
 					curr_pos = next_pos;
@@ -642,14 +750,11 @@ rawread(void)
 				}
 				else	// named pipe not seekable
 				{
-					char *dummybuf =
-						malloc(rr.scomplen+rr.pcomplen);
+					char *dummybuf = malloc(rr.scomplen+rr.pcomplen+rr.ccomplen+rr.icomplen);
 
-					ptrverify(dummybuf,
-				        "Malloc rawlog pipe buffer failed\n");
+					ptrverify(dummybuf, "Malloc rawlog pipe buffer failed\n");
 
-					readchunk(rawfd, dummybuf,
-						rr.scomplen+rr.pcomplen);
+					readchunk(rawfd, dummybuf, rr.scomplen+rr.pcomplen+rr.ccomplen+rr.icomplen);
 
 					free(dummybuf);
 				}
@@ -670,17 +775,16 @@ rawread(void)
 
 			/*
 			** allocate space, read compressed system-level
-			** statistics and decompress
+			** metrics and decompress
 			*/
 			if ( !getrawsstat(rawfd, &sstat, rr.scomplen) )
 				cleanstop(7);
 
 			/*
 			** allocate space, read compressed process-level
-			** statistics and decompress
+			** metrics and decompress
 			*/
-			devtstat.taskall    = malloc(sizeof(struct tstat  )
-								* rr.ndeviat);
+			devtstat.taskall = malloc(sizeof(struct tstat) * rr.ndeviat);
 
 			if (rr.totproc < rr.nactproc)	// compat old raw files
 				devtstat.procall = malloc(sizeof(struct tstat *)
@@ -735,8 +839,19 @@ rawread(void)
  			devtstat.totzombie	= rr.totzomb;
 
 			/*
-			** activate the installed print-function to visualize
-			** the system- and process-level statistics
+			** allocate space, read compressed cgroup-level
+			** metrics, the pidlist and decompress
+			*/
+			if (rr.flags & RRCGRSTAT)
+			{
+				if ( !getrawcstat(rawfd, &devchain, rr.ccomplen, rr.coriglen,
+							rr.icomplen, rr.ncgroups, rr.ncgpids) )
+				cleanstop(7);
+			}
+
+			/*
+			** activate the installed print function to visualize
+			** the system metrics, process metrics and cgroup metrics
 			*/
 			sampcnt++;
 
@@ -795,8 +910,8 @@ rawread(void)
 			do
 			{
 				lastcmd = (vis.show_samp)(rr.curtime,
-				     rr.interval,
-			             &devtstat, &sstat,
+				     rr.interval, &devtstat, &sstat,
+				     devchain, rr.ncgroups, rr.ncgpids,
 			             rr.nexit, rr.noverflow, flags);
 			}
 			while (!isregular &&
@@ -809,13 +924,20 @@ rawread(void)
 			free(devtstat.procall);
 			free(devtstat.procactive);
 
+			if (rr.flags & RRCGRSTAT)
+			{
+				free(devchain->cstat);
+				free(devchain->proclist);
+				free(devchain);
+			}
+
 			switch (lastcmd)
 			{
 		   	   case MSAMPPREV:
 				if (offcur >= 2)
-					offcur-= 2;
+					offcur -= 2;
 				else
-					offcur = 0;
+					offcur  = 0;
 
 				lseek(rawfd, *(offlist+offcur), SEEK_SET);
 				break;
@@ -857,6 +979,7 @@ rawread(void)
 	return isregular;
 }
 
+
 /*
 ** read the next raw record from the raw logfile
 */
@@ -866,6 +989,7 @@ getrawrec(int rawfd, struct rawrecord *prr, int rrlen)
 	return readchunk(rawfd, prr, rrlen);
 }
 
+
 /*
 ** read the system-level statistics from the current offset
 */
@@ -895,6 +1019,7 @@ getrawsstat(int rawfd, struct sstat *sp, int complen)
 	return 1;
 }
 
+
 /*
 ** read the process-level statistics from the current offset
 */
@@ -924,6 +1049,76 @@ getrawtstat(int rawfd, struct tstat *pp, int complen, int ndeviat)
 	return 1;
 }
 
+
+/*
+** read the cgroup-level statistics and pidlist from the current offset
+*/
+static int
+getrawcstat(int rawfd, struct cgchainer **cpp,
+		unsigned long ccomplen, unsigned long coriglen,
+		unsigned long icomplen, int ncgroups, int npids)
+{
+	Byte		*ccompbuf, *corigbuf, *icompbuf, *iorigbuf;
+	int		rv;
+	unsigned long	ioriglen = npids * sizeof(pid_t);
+
+	/*
+	** read all cstat structs
+	*/
+	ccompbuf = malloc(ccomplen);
+	corigbuf = malloc(coriglen);
+
+	ptrverify(ccompbuf, "Malloc failed for reading compressed cgroups\n");
+	ptrverify(corigbuf, "Malloc failed for decompressing cgroups\n");
+
+	if ( readchunk(rawfd, ccompbuf, ccomplen) < ccomplen)
+	{
+		free(ccompbuf);
+		free(corigbuf);
+		return 0;
+	}
+
+	rv = uncompress((Byte *)corigbuf, &coriglen, ccompbuf, ccomplen);
+
+	testcompval(rv, "uncompress cgroups");
+
+	free(ccompbuf);
+
+	/*
+	** read pidlist
+	*/
+	icompbuf = malloc(icomplen);
+	iorigbuf = malloc(ioriglen);
+
+	ptrverify(icompbuf, "Malloc failed for reading pidlist\n");
+	ptrverify(iorigbuf, "Malloc failed for decompresssed pidlist\n");
+
+	if ( readchunk(rawfd, icompbuf, icomplen) < icomplen)
+	{
+		free(corigbuf);
+		free(icompbuf);
+		free(iorigbuf);
+
+		return 0;
+	}
+
+	rv = uncompress((Byte *)iorigbuf, &ioriglen, icompbuf, icomplen);
+
+	testcompval(rv, "uncompress cgroups pidlist");
+
+	free(icompbuf);
+
+	/*
+	** reconstruct an array of cgchainer structs from which
+	** each entry refers to one cstat struct and its own start
+	** entry in the pidlist
+	*/
+	cgbuildarray(cpp, (char *)corigbuf, (char *)iorigbuf, ncgroups);
+
+	return 1;
+}
+
+
 /* 
 ** verify if a particular ascii-string is in the format yyyymmdd
 */
@@ -1052,5 +1247,6 @@ try_other_version(int majorversion, int minorversion)
 	/*
 	** point of no return, except when exec failed
 	*/
-	fprintf(stderr, "activation of %s failed!\n", tmpbuf);
+	fprintf(stderr, "activation of %s failed!\n\n", tmpbuf);
+	fprintf(stderr, "use 'atopconvert' to convert this raw log!\n");
 }
